"""
Main entry point for `routemaster` command-line utility.

This is to enable `python -m routemaster` if that is needed for any reason,
normal use should be to use the `routemaster` command-line tool directly.
"""

from routemaster.cli import main

main()
